#  MeepMeep: fast orbit calculations for exoplanet modelling
#  Copyright (C) 2022 Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from numba import njit
from numpy import pi, arctan2, sqrt, sin, cos, arccos, mod, copysign, sign, array, arcsin
from scipy.constants import G

HALF_PI = 0.5*pi
TWO_PI = 2.0*pi


@njit
def eccentricity_vector(i, e, w):
    if e > 1e-5:
        ci = cos(i)
        si = sin(i)
        ex = -e*cos(w)
        ey = -e*sin(w)*ci
        ez =  e*sin(w)*si
        return array([ex, ey, ez])
    else:
        return array([-1.0, 0.0, 0.0])


@njit
def eclipse_phase(p, i, e, w):
    """ Phase for the secondary eclipse center.

    Exact secondary eclipse center phase, good for all eccentricities.
    """
    etr = arctan2(sqrt(1. - e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
    eec = arctan2(sqrt(1. - e**2) * sin(HALF_PI + pi - w), e + cos(HALF_PI + pi - w))
    mtr = etr - e * sin(etr)
    mec = eec - e * sin(eec)
    phase = (mec - mtr) * p / TWO_PI
    return phase if phase > 0. else p + phase


@njit
def af_transit(e, w):
    """Calculates the -- factor during the transit"""
    return (1.0-e**2)/(1.0 + e*sin(w))


@njit
def i_from_baew(b, a, e, w):
    """Orbital inclination from the impact parameter, scaled semi-major axis, eccentricity and argument of periastron

    Parameters
    ----------

      b  : impact parameter       [-]
      a  : scaled semi-major axis [R_Star]
      e  : eccentricity           [-]
      w  : argument of periastron [rad]

    Returns
    -------

      i  : inclination            [rad]
    """
    return arccos(b / (a*af_transit(e, w)))


@njit
def as_from_rhop(rho, period):
    """Scaled semi-major axis from the stellar density and planet's orbital period.

    Parameters
    ----------

      rho    : stellar density [g/cm^3]
      period : orbital period  [d]

    Returns
    -------

      as : scaled semi-major axis [R_star]
    """
    return (G/(3*pi))**(1/3) * ((period * 86400.0)**2 * 1e3 * rho)**(1 / 3)


@njit
def ta_from_ea_v(Ea, e):
    sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
    cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
    Ta  = arctan2(sta, cta)
    return Ta


@njit
def ta_from_ea_s(Ea, e):
    sta = sqrt(1.0-e**2) * sin(Ea)/(1.0-e*cos(Ea))
    cta = (cos(Ea)-e)/(1.0-e*cos(Ea))
    Ta  = arctan2(sta, cta)
    return Ta


@njit
def mean_anomaly_offset(e, w):
    mean_anomaly_offset = arctan2(sqrt(1.0-e**2) * sin(HALF_PI - w), e + cos(HALF_PI - w))
    mean_anomaly_offset -= e*sin(mean_anomaly_offset)
    return mean_anomaly_offset


@njit
def mean_anomaly(t, t0, p, e, w):
    offset = mean_anomaly_offset(e, w)
    Ma = mod(TWO_PI * (t - (t0 - offset * p / TWO_PI)) / p, TWO_PI)
    return Ma


@njit
def z_from_ta_s(Ta, a, i, e, w):
    z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
    z *= copysign(1.0, sin(w+Ta))
    return z


@njit(parallel=True)
def z_from_ta_v(Ta, a, i, e, w):
    z  = a*(1.0-e**2)/(1.0+e*cos(Ta)) * sqrt(1.0 - sin(w+Ta)**2 * sin(i)**2)
    z *= sign(1.0, sin(w+Ta))
    return z


@njit
def impact_parameter(a, i):
    return a * cos(i)


@njit
def impact_parameter_ec(a, i, e, w, tr_sign):
    return a * cos(i) * ((1.-e**2) / (1.+tr_sign*e*sin(w)))

@njit
def d_from_pkaiews(p, k, a, i, e, w, tr_sign, kind=14):
    """Transit duration (T14 or T23) from p, k, a, i, e, w, and the transit sign.

    Calculates the transit duration (T14) from the orbital period, planet-star radius ratio, scaled semi-major axis,
    orbital inclination, eccentricity, argument of periastron, and the sign of the transit (transit:1, eclipse: -1).

     Parameters
     ----------

       p  : orbital period         [d]
       k  : radius ratio           [R_Star]
       a  : scaled semi-major axis [R_star]
       i  : orbital inclination    [rad]
       e  : eccentricity           [-]
       w  : argument of periastron [rad]
       tr_sign : transit sign, 1 for a transit, -1 for an eclipse
       kind: either 14 for full transit duration or 23 for total transit duration

     Returns
     -------

       d  : transit duration T14  [d]
     """
    b  = impact_parameter_ec(a, i, e, w, tr_sign)
    ae = sqrt(1.-e**2)/(1.+tr_sign*e*sin(w))
    ds = 1. if kind == 14 else -1.
    return p/pi  * arcsin(sqrt((1.+ds*k)**2-b**2)/(a*sin(i))) * ae


from numba import njit
from numpy import cos, sin, zeros

from .utils import mean_anomaly, ta_from_ea_s, ta_from_ea_v, z_from_ta_s, z_from_ta_v, eclipse_phase


@njit
def ea_newton_s(t, t0, p, e, w):
    ma = mean_anomaly(t, t0, p, e, w)
    ea = ma
    err = 0.05
    k = 0
    while abs(err) > 1e-8 and k < 1000:
        err = ea - e*sin(ea) - ma
        ea = ea - err/(1.0-e*cos(ea))
        k += 1
    return ea


@njit
def ea_newton_v(t, t0, p, e, w):
    ea = zeros(t.size)
    for i in range(len(t)):
        ea[i] = ea_newton_s(t[i], t0, p, e, w)
    return ea


@njit
def ta_newton_s(t, t0, p, e, w):
    return ta_from_ea_s(ea_newton_s(t, t0, p, e, w), e)


@njit
def ta_newton_v(t, t0, p, e, w):
    return ta_from_ea_v(ea_newton_v(t, t0, p, e, w), e)


@njit(fastmath=True)
def xy_newton_v(time, t0, p, a, i, e, w):
    """Planet velocity and acceleration at mid-transit in [R_star / day]"""
    f = ta_newton_v(time, t0, p, e, w)
    r = a * (1. - e ** 2) / (1. + e * cos(f))
    x = -r * cos(w + f)
    y = -r * sin(w + f) * cos(i)
    return x, y


@njit(fastmath=True)
def xyz_newton_v(time, t0, p, a, i, e, w):
    """Planet velocity and acceleration at mid-transit in [R_star / day]"""
    f = ta_newton_v(time, t0, p, e, w)
    r = a * (1. - e ** 2) / (1. + e * cos(f))
    x = -r * cos(w + f)
    y = -r * sin(w + f) * cos(i)
    z =  r * sin(w + f) * sin(i)
    return x, y, z


@njit
def z_newton_s(time, t0, p, a, i, e, w):
    """Normalized projected distance for scalar time.
    """
    ta = ta_newton_s(time, t0, p, e, w)
    return z_from_ta_s(ta, a, i, e, w)


@njit
def z_newton_v(time, t0, p, a, i, e, w):
    """Normalized projected distance for an array of times.
    """
    ta = ta_newton_v(time, t0, p, e, w)
    return z_from_ta_v(ta, a, i, e, w)


@njit
def rv_newton_v(times, k, t0, p, e, w):
    ta_n = ta_newton_v(times, t0, p, e, w)
    return k * (cos(w + ta_n) + e * cos(w))


@njit
def eclipse_light_travel_time(p: float, a: float, i: float, e: float, w: float, rstar: float):
    """
    Calculate the light travel time difference between the transit and the secondary eclipse of an exoplanet.

    This function computes the difference in light travel time caused by the displacement of the planet between
    its transit across the star and its secondary eclipse (when the planet passes behind the star as viewed from Earth).

    Parameters
    ----------
    p : float
        Orbital period in days.
    a : float
        Semi-major axis of the planet's orbit in host star radii.
    i : float
        Orbital inclination in radians.
    e : float
        Orbital eccentricity.
    w : float
        Argument of periastron in radians.
    rstar : float
        Radius of the star in solar radii.

    Returns
    -------
    float
        The light travel time difference in days between the transit and secondary eclipse of the exoplanet.
    """
    s = 2.685885891543453e-05  # Light travel time for a distance of one solar radius in days

    ae = a * (1. - e ** 2)
    si = sin(i)

    f = ta_newton_s(0.0, 0.0, p, e, w)
    r = ae / (1. + e * cos(f))
    ztr = r * sin(w + f) * si

    f = ta_newton_s(eclipse_phase(p, i, e, w), 0.0, p, e, w)
    r = ae / (1. + e * cos(f))
    zec = r * sin(w + f) * si

    return (ztr - zec) * rstar * s


#  MeepMeep: fast orbit calculations for exoplanet modelling
#  Copyright (C) 2022 Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit
from numpy import zeros_like, cos, sin, floor, sqrt, zeros, linspace
from .utils import eclipse_phase, i_from_baew
from .newton import ta_newton_s, ta_newton_v


@njit(fastmath=True)
def solve_xy5s(phase, p, a, i, e, w):
    """Planet velocity, acceleration, jerk, and snap at mid-transit in [R_star / day]"""

    # Time step for central finite difference
    # ---------------------------------------
    # I've tried to choose a value that is small enough to
    # work with ultra-short-period orbits and large enough
    # not to cause floating point problems with the fourth
    # derivative (anything much smaller starts hitting the
    # double precision limit.)
    dt = 2e-2

    ae = a*(1. - e**2)
    ci = cos(i)

    # Calculation of X and Y positions
    # --------------------------------
    # These could just as well be calculated with a single
    # loop with X and Y as arrays, but I've decided to
    # manually unroll it because it seems to give a small
    # speed advantage with numba.

    f0 = ta_newton_s(phase-3 * dt, 0.0, p, e, w)
    f1 = ta_newton_s(phase-2 * dt, 0.0, p, e, w)
    f2 = ta_newton_s(phase-dt, 0.0, p, e, w)
    f3 = ta_newton_s(phase, 0.0, p, e, w)
    f4 = ta_newton_s(phase+dt, 0.0, p, e, w)
    f5 = ta_newton_s(phase+2 * dt, 0.0, p, e, w)
    f6 = ta_newton_s(phase+3 * dt, 0.0, p, e, w)

    r0 = ae/(1. + e*cos(f0))
    r1 = ae/(1. + e*cos(f1))
    r2 = ae/(1. + e*cos(f2))
    r3 = ae/(1. + e*cos(f3))
    r4 = ae/(1. + e*cos(f4))
    r5 = ae/(1. + e*cos(f5))
    r6 = ae/(1. + e*cos(f6))

    x0 = -r0*cos(w + f0)
    x1 = -r1*cos(w + f1)
    x2 = -r2*cos(w + f2)
    x3 = -r3*cos(w + f3)
    x4 = -r4*cos(w + f4)
    x5 = -r5*cos(w + f5)
    x6 = -r6*cos(w + f6)

    y0 = -r0*sin(w + f0)*ci
    y1 = -r1*sin(w + f1)*ci
    y2 = -r2*sin(w + f2)*ci
    y3 = -r3*sin(w + f3)*ci
    y4 = -r4*sin(w + f4)*ci
    y5 = -r5*sin(w + f5)*ci
    y6 = -r6*sin(w + f6)*ci

    # First time derivative of position: velocity
    # -------------------------------------------
    a, b, c = 1/60, 9/60, 45/60
    vx = (a*(x6 - x0) + b*(x1 - x5) + c*(x4 - x2))/dt
    vy = (a*(y6 - y0) + b*(y1 - y5) + c*(y4 - y2))/dt

    # Second time derivative of position: acceleration
    # ------------------------------------------------
    a, b, c, d = 1/90, 3/20, 3/2, 49/18
    ax = (a*(x0 + x6) - b*(x1 + x5) + c*(x2 + x4) - d*x3)/dt**2
    ay = (a*(y0 + y6) - b*(y1 + y5) + c*(y2 + y4) - d*y3)/dt**2

    # Third time derivative of position: jerk
    # ---------------------------------------
    a, b, c = 1/8, 1, 13/8
    jx = (a*(x0 - x6) + b*(x5 - x1) + c*(x2 - x4))/dt**3
    jy = (a*(y0 - y6) + b*(y5 - y1) + c*(y2 - y4))/dt**3

    # Fourth time derivative of position: snap
    # ----------------------------------------
    a, b, c, d = 1/6, 2, 13/2, 28/3
    sx = (-a*(x0 + x6) + b*(x1 + x5) - c*(x2 + x4) + d*x3)/dt**4
    sy = (-a*(y0 + y6) + b*(y1 + y5) - c*(y2 + y4) + d*y3)/dt**4

    return x3, y3, vx, vy, ax, ay, jx, jy, sx, sy





@njit(fastmath=True)
def prdist_5s(tc, t0, p, x0, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    """Normalized planet-star center distance using Taylor series expansion.
    """
    epoch = floor((tc - t0 + 0.5 * p) / p)
    t = tc - (t0 + epoch * p)
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    px = x0 + vx * t + 0.5 * ax * t2 + jx * t3 / 6.0 + sx * t4 / 24.
    py = y0 + vy * t + 0.5 * ay * t2 + jy * t3 / 6.0 + sy * t4 / 24.
    return sqrt(px ** 2 + py ** 2)


@njit(fastmath=True)
def prdist_4s(tc, t0, p, x0, y0, vx, vy, ax, ay, jx, jy):
    """Normalized planet-star center distance using Taylor series expansion.
    """
    epoch = floor((tc - t0 + 0.5 * p) / p)
    t = tc - (t0 + epoch * p)
    t2 = t * t
    t3 = t2 * t
    px = x0 + vx * t + 0.5 * ax * t2 + jx * t3 / 6.0
    py = y0 + vy * t + 0.5 * ay * t2 + jy * t3 / 6.0
    return sqrt(px ** 2 + py ** 2)


@njit(fastmath=True)
def z_taylor_v(times, t0, p, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    z = zeros_like(times)
    npt = times.size
    for i in range(npt):
        epoch = floor((times[i] - t0 + 0.5 * p) / p)
        t = times[i] - (t0 + epoch * p)
        t2 = t * t
        t3 = t2 * t
        t4 = t3 * t
        px =      vx * t + 0.5 * ax * t2 + jx * t3 / 6.0 + sx * t4 / 24.
        py = y0 + vy * t + 0.5 * ay * t2 + jy * t3 / 6.0 + sy * t4 / 24.
        z[i] = sqrt(px ** 2 + py ** 2)
    return z


@njit(fastmath=True)
def xy_newton_v(times, t0, p, a, b, e, w):
    """Planet velocity and acceleration at mid-transit in [R_star / day]"""
    i = i_from_baew(b, a, e, w)
    f = ta_newton_v(times, t0, p, e, w)
    r = a * (1. - e ** 2) / (1. + e * cos(f))
    x = -r * cos(w + f)
    y = -r * sin(w + f) * cos(i)
    return x, y


@njit(fastmath=True)
def xy_taylor_v(times, t0, p, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    npt = times.size
    px = zeros(npt)
    py = zeros(npt)
    for i in range(npt):
        epoch = floor((times[i] - t0 + 0.5 * p) / p)
        t = times[i] - (t0 + epoch * p)
        px[i] =      vx * t + 0.5 * ax * t ** 2 + jx * t ** 3 / 6. + sx * t**4 / 24.
        py[i] = y0 + vy * t + 0.5 * ay * t ** 2 + jy * t ** 3 / 6. + sy * t**4 / 24.
    return px, py


@njit
def find_contact_point(k: float, point: int, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    if point == 1 or point == 2 or point == 12:
        s = -1.0
    else:
        s = 1.0

    if point == 1 or point == 4:
        zt = 1.0 + k
    elif point == 2 or point == 3:
        zt = 1.0 - k
    else:
        zt = 1.0

    t0 = 0.0
    t2 = s*2.0/vx
    t1 = 0.5*t2

    z0 = prdist_5s(t0, 0.0, 1.0, y0, vx, vy, ax, ay, jx, jy, sx, sy) - zt
    z1 = prdist_5s(t1, 0.0, 1.0, y0, vx, vy, ax, ay, jx, jy, sx, sy) - zt

    i = 0
    while abs(t2 - t0) > 1e-6 and i < 100:
        if z0*z1 < 0.0:
            t1, t2 = 0.5*(t0 + t1), t1
            z1, z2 = prdist_5s(t1, 0.0, 1.0, y0, vx, vy, ax, ay, jx, jy, sx, sy) - zt, z1
        else:
            t0, t1 = t1, 0.5*(t1 + t2)
            z0, z1 = z1, prdist_5s(t1, 0.0, 1.0, y0, vx, vy, ax, ay, jx, jy, sx, sy) - zt
        i += 1
    return t1


@njit
def find_z_min(tc, p, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    r = 0.61803399
    c = 1.0 - r
    x0, x3 = tc - 0.01, tc + 0.01

    x1 = tc
    x2 = tc + c*(x3 - tc)

    f1 = prdist_5s(x1, tc, p, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    f2 = prdist_5s(x2, tc, p, y0, vx, vy, ax, ay, jx, jy, sx, sy)

    i = 0
    while abs(x3 - x0) > 1e-7 and i < 100:
        if f2 < f1:
            x0, x1, x2 = x1, x2, r*x2 + c*x3
            f1, f2 = f2, prdist_5s(x2, tc, p, y0, vx, vy, ax, ay, jx, jy, sx, sy)
        else:
            x3, x2, x1 = x2, x1, r*x1 + c*x0
            f2, f1 = f1, prdist_5s(x1, tc, p, y0, vx, vy, ax, ay, jx, jy, sx, sy)
        i += 1

    if f1 < f2:
        return x1, f1
    else:
        return x2, f2

@njit
def t14(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t1 = find_contact_point(k, 1, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t4 = find_contact_point(k, 4, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    return t4-t1


def t23(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t2 = find_contact_point(k, 2, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t3 = find_contact_point(k, 3, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    return t3-t2


def t12(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t1 = find_contact_point(k, 1, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t2 = find_contact_point(k, 2, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    return t2-t1


def t34(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t3 = find_contact_point(k, 3, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t4 = find_contact_point(k, 4, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    return t4-t3


def t1(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    return find_contact_point(k, 1, y0, vx, vy, ax, ay, jx, jy, sx, sy)


def t4(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    return find_contact_point(k, 4, y0, vx, vy, ax, ay, jx, jy, sx, sy)


def bounding_box(k: float, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t1 = find_contact_point(k, 1, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t4 = find_contact_point(k, 4, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    return t1, t4


from math import isfinite

from numba import njit
from numpy import zeros, sqrt, nan

from .position import solve_xy_p5s, xy_t15s, xy_t15sc


@njit
def xy_derivative_coeffs(phase, p, a, i, e, w, pds, c0, eps):
    cs = zeros((6, 2, 5))
    for j in range(6):
        v = pds[j]*eps
        cs[j, :, :] = solve_xy_p5s(phase+v[0], p+v[1], a+v[2], i+v[3], e+v[4], w+v[5])
    return (cs - c0) / eps


@njit
def loc_and_der_coeffs(phase, p, a, i, e, w, diffs):
    """Expands a planet's (x,y) location into a Taylor series.

    Expands a planet's (x,y) location into a Taylor series and optionally also expands the
    location's partial derivatives.
    """
    nor = p.size

    # ---------------------------------------------------------------------
    # Solve the Taylor series coefficients for the planet's (x, y) location
    # and its derivatives if they're requested.
    # ---------------------------------------------------------------------
    cfs = zeros((nor, 2, 5))
    for j in range(nor):
        if a[j] > 1.0 and e[j] < 0.99:
            cfs[j, :, :] = solve_xy_p5s(phase, p[j], a[j], i[j], e[j], w[j])
        else:
            cfs[j, :, :] = nan

    if diffs is not None:
        dcfs = zeros((nor, 6, 2, 5))
        for j in range(nor):
            if isfinite(cfs[j, 0, 0]):
                dcfs[j, :, :, :] = xy_derivative_coeffs(phase, p[j], a[j], i[j], e[j], w[j],
                                                        diffs[j], cfs[j], 1e-4)
            else:
                dcfs[j, :, :, :] = nan
    else:
        dcfs = zeros((0, 0, 0, 0))
    return cfs, dcfs

# Position derivatives
# --------------------

@njit
def dxy_dtc(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[0])


@njit
def dxy_dp(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[1])


@njit
def dxy_da(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[2])


@njit
def dxy_di(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[3])


@njit
def dxy_de(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[4])


@njit
def dxy_dw(t, t0, p, dcs):
    return xy_t15s(t, t0, p, dcs[5])


# Projected distance derivatives
# ------------------------------
@njit(fastmath=True)
def dpd(t, x, y, dcs):
    dx, dy = xy_t15sc(t, dcs)
    return (0.5/sqrt(x**2 + y**2))*(2*x*dx + 2*y*dy)


@njit(fastmath=True)
def pd_derivatives_s(t, x, y, dcf, res):
    res[0] = dpd(t, x, y, dcf[0])  # 0: Zero epoch
    res[1] = dpd(t, x, y, dcf[1])  # 1: Period
    res[2] = dpd(t, x, y, dcf[2])  # 2: Semi-major axis
    res[3] = dpd(t, x, y, dcf[3])  # 3: Inclination
    res[4] = dpd(t, x, y, dcf[4])  # 4: Eccentricity
    res[5] = dpd(t, x, y, dcf[5])  # 5: Argument of periastron
    return res


@njit(fastmath=True)
def pd_with_derivatives_s(t, cf, dcf, res):
    x, y = xy_t15sc(t, cf)
    res[0] = sqrt(x**2 + y**2)        # 0: Projected distance [R_Sun]
    res[1] = dpd(t, x, y, dcf[0])     # 1: Zero epoch
    res[2] = dpd(t, x, y, dcf[1])     # 2: Period
    res[3] = dpd(t, x, y, dcf[2])     # 3: Semi-major axis
    res[4] = dpd(t, x, y, dcf[3])     # 4: Inclination
    res[5] = dpd(t, x, y, dcf[4])     # 5: Eccentricity
    res[6] = dpd(t, x, y, dcf[5])  # 6: Argument of periastron
    return res


@njit
def pd_with_derivatives_v(t, cf, dcf):
    npt = t.size
    res = zeros((7, npt))
    for i in range(npt):
        pd_with_derivatives_s(t[i], cf, dcf, res[:, i])
    return res


@njit
def dpd_dtc(t, x, y, dcs):
    return dpd(t, x, y, dcs[0])


@njit
def dpd_dp(t, p, xy, dcs):
    return dpd(p, xy, t, dcs[1])


@njit
def dpd_da(t, p, xy, dcs):
    return dpd(p, xy, t, dcs[2])


@njit
def dpd_di(t, p, xy, dcs):
    return dpd(p, xy, t, dcs[3])


@njit
def dpd_de(t, p, xy, dcs):
    return dpd(p, xy, t, dcs[4])


@njit
def dpd_dw(t, p, xy, dcs):
    return dpd(p, xy, t, dcs[5])
